<?php

namespace Drupal\gcs_backup\Commands;

use Drupal\gcs_backup\GcsBackupManager;
use Drush\Commands\DrushCommands;
use Symfony\Component\Console\Helper\Table;

/**
 * Google cloud Backup commands.
 */
class BackupCommands extends DrushCommands {

  /**
   * Backup manager.
   *
   * @var \Drupal\gcs_backup\GcsBackupManager
   */
  private GcsBackupManager $gcsBackupManager;

  public function __construct(
    GcsBackupManager $gcs_backup_manager,
  ) {
    parent::__construct();

    $this->gcsBackupManager = $gcs_backup_manager;
  }

  /**
   * List backups and download selected one.
   *
   * @command gsc:list-download
   * @aliases gscld
   */
  public function listAndDownload() {

    $backups = $this->gcsBackupManager->getBackups();
    if (empty($backups)) {
      $this->output()->writeln('No backups found.');
      return;
    }

    $table = new Table($this->output());
    $table->setHeaders(['Index', 'Filename', 'Size', 'Date']);

    foreach ($backups as $index => $object) {
      $info = $object->info();

      $size = $info['size'] ?? '';
      $date = $info['updated'] ?? '';

      $table->addRow([
        $index + 1,
        $object->name(),
        $size,
        $date,
      ]);
    }

    $table->render();

    // @phpstan-ignore-next-line
    $choice = $this->io()->ask('Enter the index of the backup to download (or 0 to cancel): ');

    if ($choice > 0 && $choice <= count($backups)) {
      $selectedBackup = $backups[$choice - 1];
      try {
        $destination = $this->gcsBackupManager->downloadBackup($selectedBackup->name());
      }
      catch (\Exception $e) {
        $this->output()->writeln('Error on backup download. Error: ' . $e->getMessage());
        return;
      }

      if (!empty($destination)) {
        $this->output()->writeln("Backup downloaded to: $destination");
      }
      else {
        $this->output()->writeln('No backup downloaded.');
      }

    }
    elseif ($choice != 0) {
      $this->output()->writeln('Invalid choice. No backup downloaded.');
    }
  }

  /**
   * Create backup and upload to Google Cloud.
   *
   * @command gsc:create-bk
   * @aliases gsccbk
   */
  public function createBackup() {
    try {
      $file_uri = $this->gcsBackupManager->createBackup();
      $this->gcsBackupManager->uploadBackupToGcs($file_uri);
    }
    catch (\Exception $e) {
      $this->output()->writeln($e->getMessage());
      return;
    }
    $this->output()->writeln('Backup created and uploaded successfully.');
  }

}
